"""
Dune source for data extraction via REST API.

This source provides access to Dune blockchain analytics data via SQL query execution.
"""

from typing import Any, Iterator

import dlt

from ingestr.src.dune.helpers import (
    BASE_URL,
    fetch_queries,
    fetch_results,
    poll_execution,
)
from ingestr.src.http_client import create_client


@dlt.source(max_table_nesting=0, name="dune_source")
def dune_source(
    api_key: str,
    sql: str | None = None,
    query_id: str | None = None,
    performance: str = "medium",
    query_parameters: dict[str, Any] | None = None,
) -> Any:
    """
    Dune data source for blockchain analytics data extraction.

    Args:
        api_key: Dune API key for authentication
        sql: The SQL query to execute
        query_id: Optional saved query ID to execute via /query/{id}/execute
        performance: Engine tier - "medium" (default) or "large"
        query_parameters: Optional key-value pairs for parameterized queries
    """
    session = create_client()
    headers = {"X-DUNE-API-KEY": api_key}

    @dlt.resource(
        name="execute_sql",
        write_disposition="replace",
    )
    def execute_sql() -> Iterator[dict[str, Any]]:
        execute_payload: dict[str, Any] = {
            "sql": sql,
            "performance": performance,
        }

        execute_response = session.post(
            f"{BASE_URL}/sql/execute",
            json=execute_payload,
            headers=headers,
        )
        execute_response.raise_for_status()
        execute_data = execute_response.json()

        execution_id = execute_data.get("execution_id")
        if not execution_id:
            raise ValueError(f"Failed to start query execution: {execute_data}")

        poll_execution(session, headers, execution_id)
        yield from fetch_results(session, headers, execution_id)

    @dlt.resource(
        name="execute_query",
        write_disposition="replace",
    )
    def execute_query() -> Iterator[dict[str, Any]]:
        execute_payload: dict[str, Any] = {
            "performance": performance,
        }
        if query_parameters:
            execute_payload["query_parameters"] = query_parameters

        execute_response = session.post(
            f"{BASE_URL}/query/{query_id}/execute",
            json=execute_payload,
            headers=headers,
        )
        execute_response.raise_for_status()
        execute_data = execute_response.json()

        execution_id = execute_data.get("execution_id")
        if not execution_id:
            raise ValueError(f"Failed to start query execution: {execute_data}")

        poll_execution(session, headers, execution_id)
        yield from fetch_results(session, headers, execution_id)

    @dlt.resource(
        name="queries",
        write_disposition="replace",
        primary_key="id",
    )
    def queries() -> Iterator[dict[str, Any]]:
        yield from fetch_queries(session, headers)

    if query_id:
        return execute_query
    if sql:
        return execute_sql
    return (queries,)



import time
from typing import Any, Iterator

BASE_URL = "https://api.dune.com/api/v1"


def poll_execution(session, headers: dict, execution_id: str) -> None:
    max_retries = 8640  # Max 12 hours with 5-second intervals
    retry_count = 0
    poll_interval = 5

    while retry_count < max_retries:
        status_response = session.get(
            f"{BASE_URL}/execution/{execution_id}/status",
            headers=headers,
        )
        status_response.raise_for_status()
        status_data = status_response.json()
        state = status_data.get("state")

        if state == "QUERY_STATE_COMPLETED":
            return
        elif state == "QUERY_STATE_FAILED":
            error = status_data.get("error", {})
            error_msg = (
                error.get("message", "Unknown error")
                if isinstance(error, dict)
                else str(error)
            )
            raise ValueError(f"Query execution failed: {error_msg}")
        elif state in ("QUERY_STATE_PENDING", "QUERY_STATE_EXECUTING"):
            time.sleep(poll_interval)
            retry_count += 1
        elif state == "QUERY_STATE_CANCELLED":
            raise ValueError("Query execution was cancelled")
        elif state == "QUERY_STATE_EXPIRED":
            raise ValueError("Query execution expired")
        else:
            raise ValueError(f"Unknown query state: {state}")

    raise TimeoutError(
        f"Query execution timed out after {max_retries * poll_interval} seconds"
    )


def fetch_results(
    session, headers: dict, execution_id: str
) -> Iterator[dict[str, Any]]:
    offset = 0
    page_limit = 1000

    while True:
        params: dict[str, Any] = {
            "limit": page_limit,
            "offset": offset,
        }

        results_response = session.get(
            f"{BASE_URL}/execution/{execution_id}/results",
            headers=headers,
            params=params,
        )
        results_response.raise_for_status()
        results_data = results_response.json()

        result = results_data.get("result", {})
        rows = result.get("rows", [])

        if not rows:
            break

        yield from rows

        next_offset = results_data.get("next_offset")
        if not next_offset:
            break

        offset = next_offset


def fetch_queries(session, headers: dict) -> Iterator[dict[str, Any]]:
    offset = 0
    page_limit = 100

    while True:
        params: dict[str, Any] = {
            "limit": page_limit,
            "offset": offset,
        }

        response = session.get(
            f"{BASE_URL}/queries",
            headers=headers,
            params=params,
        )
        response.raise_for_status()
        data = response.json()

        rows = data.get("queries", [])
        if not rows:
            break

        yield from rows

        total = data.get("total", 0)
        offset += len(rows)
        if offset >= total:
            break
